require 'rubygems'

gem 'sinatra'
gem 'grit'
gem 'rmagick'
gem 'mime-types'

require 'RMagick'
require 'sinatra'
require 'grit'
require 'mime/types'
require 'yaml'
require 'delegate'
require 'pathname'
require 'fileutils'

include Magick

CONFIG_PATH = "config.yml"
DEFAULT_RENDERER = "haml"
DOWNLOAD_LINK = "http://github.com/erikh/photorepo"

# widths, aspect ratio is preserved for height.
SIZE_DEFAULTS = {
    :thumb  => 150,
    :small  => 300,
    :medium => 600,
    :large  => 1200
}

module Photo
    class Config < DelegateClass(Hash)
        def initialize(file)
            super(YAML.load_file(file))
        end

        def [](name)
            super(name.to_s)
        end

        def []=(name, value)
            super(name.to_s, value)
        end
    end

    class File

        attr_reader :path
        attr_reader :name
        attr_reader :data

        def initialize(path, data)
            @path = path
            @name = path.basename
            @data = data
        end

        def repo_path
            dirname, basename = Pathname.new(@path).split
            dirname.basename + basename
        end

        def to_str
            path.to_s
        end
    end

    class Collection

        include Enumerable

        attr_reader :name

        def initialize(name, photo_repo, create=false)
            @name = name
            @repo = photo_repo

            @tree = @repo / name
            @path = @repo.repo_path + name
        end

        def load(filename)
            self[filename]
        end

        def /(filename)
            load(filename)
        end

        def +(filename)
            pname = Pathname.new('/' + name)
            case filename
            when String
                pname + filename
            when File
                pname + filename.name
            end
        end

        def each
            @tree.blobs.each do |blob|
                yield File.new(@path + blob.name, blob.data)
            end
        end

        def to_s
            @name
        end

        def [](req_name)
            req_name = Pathname.new(req_name)
            find { |file| file.name == req_name }
        end
    end

    class Repo
        attr_reader :repo
        attr_reader :repo_path

        include Enumerable

        def initialize(repo_path)
            @repo_path = Pathname.new(repo_path)
            @repo = Grit::Repo.new(repo_path)
        end

        def [](collection)
            Collection.new(collection, self)
        end

        def /(name)
            @repo.tree / name
        end

        def each
            @repo.tree.trees.each do |tree|
                yield Collection.new(tree.name, self)
            end
        end
    end

    class Fetch

        attr_accessor :resize
        attr_accessor :rotate

        def initialize(file)
            @file = file
            @resize = nil
            @rotate = nil
            @image  = nil
        end

        def format
            MIME::Types.type_for(fetch).first.content_type
        end

        def fetch
            return locate_cache || cache
        end

        protected

        def manipulate_image
            create_image
            resize_image if @resize
            rotate_image if @rotate
        end

        def create_image
            @image ||= Image.from_blob(@file.data)[0]
        end

        def resize_image
            if IMAGE_SIZES[@resize]
                @image = @image.resize_to_fit(IMAGE_SIZES[@resize])
            else
                @resize = nil
            end
        end

        def rotate_image
            if (@rotate % 90) == 0
                @image = @image.rotate(@rotate)
            else
                @rotate = nil
            end
        end

        def build_cache_filename
            rotate_part = @rotate || 'none'
            resize_part = @resize || 'none'
            repo_path = @file.repo_path
            file = repo_path.basename(repo_path.extname)
            extension = repo_path.extname.sub(/^\./, '')
            repo_path = repo_path.dirname + file
            [repo_path.to_s, rotate_part, resize_part, extension].join('.')
        end

        def cache_filename_path(fn)
            ::File.join(CONFIG[:cache], fn)
        end

        def locate_cache
            fn = build_cache_filename
            path = cache_filename_path(fn)
            return path if fn and ::File.exists? path
            return false
        end

        def cache
            manipulate_image

            fn = build_cache_filename
            path = cache_filename_path(fn)

            unless ::File.exists?(::File.dirname(path))
                FileUtils.mkdir_p(::File.dirname(path))
            end

            ::File.open(path, 'w') { |f| f << @image.to_blob }

            return path
        end
    end
end

CONFIG = Photo::Config.new(Pathname.new(CONFIG_PATH))
REPO   = Photo::Repo.new(CONFIG[:repo])

if CONFIG[:sizes]
    hsh = CONFIG[:sizes]
    hsh.each_key do |key|
        hsh[key.to_sym] = hsh.delete(key)
    end
    IMAGE_SIZES = hsh
else
    IMAGE_SIZES = SIZE_DEFAULTS
end

CONFIG[:renderer] ||= DEFAULT_RENDERER

begin
    gem CONFIG[:renderer]
rescue Gem::LoadError
end

require CONFIG[:renderer]

class Sinatra::Application
    alias_method :myrender, CONFIG[:renderer]
end

helpers do
    def trunc(txt, size=20)
        txt = txt.to_s.dup
        return txt unless txt.length > size
        txt[(size-3)..-1] = '...'
        return txt
    end
end

get '/' do
    @route = "Home"
    @names = REPO.collect { |col| col.name }
    myrender :index
end

['/:collection', '/:collection/'].each do |route|
    get route do
        @collection = REPO[params[:collection]]
        @route      = "Collection '#{@collection.name}'"
        @files      = @collection.collect { |file| (file.name && file) rescue nil }.compact

        myrender :collection
    end
end

get '/:collection/view/:filename' do
    @collection    = REPO[params[:collection]]
    @path          = @collection / params[:filename]
    @href          = '/' + params[:collection] + '/' + params[:filename]

    unless @path 
        return status 404
    end

    @route         = "View Image"
    @information   = "File: #{@path.name}, Collection: #{@collection}"

    myrender :view_file
end

get '/:collection/:filename' do
    fetch = Photo::Fetch.new(REPO[params[:collection]] / params[:filename])
    content_type "image/#{fetch.format}"
    send_file fetch.fetch
end

[ '/:collection/:filename/resize/:size', '/:collection/:filename/resize/:size/rotate/:degree'].each do |route|
    get route do
        size = params[:size].to_sym

        unless IMAGE_SIZES[size]
            return status 404
        end

        file = REPO[params[:collection]] / params[:filename]

        fetch = Photo::Fetch.new(file)
        fetch.resize = size

        if request.path_info['rotate']
            degree = params[:degree].to_i
            fetch.rotate = degree
        end

        content_type "image/#{fetch.format}"
        send_file fetch.fetch
    end
end
